#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"

/**
 *  names.c:
 *      Deals with file names or path names as strings.
 */

void init_name(struct name *name)
{
    name->destroyed = false;
    name->prim_buf[0] = '\0';
    name->second_buf = NULL;
    name->length = 0;
    name->size = NAME_PRIMBUFSIZE;
}

bool concat_name(struct name *dest, struct name *src)
{
    assert(dest != src);
    if (dest == src) {
        return false;
    }
    return concat_name_str_part(dest, get_buffer_of_name(src),
        get_buffer_of_name(src) + src->length);
}

bool concat_name_str(struct name *name, const char *content)
{
    assert(!(get_buffer_of_name(name) <= content && content <
        get_buffer_of_name(name) + name->length));
    return concat_name_str_part(name, content, content + strlen(content));
}

char *get_buffer_of_name(struct name *name)
{
    if (name->second_buf == NULL) {
        return name->prim_buf;
    } else {
        return name->second_buf;
    }
}

/**
 *  concat_name_part:
 *      Concatenates part (substring) of a name.
 *
 *  Parameters:
 *      dest - the destination name to copy to
 *      src - the source name
 *      begin - beginning position of the part to copy
 *      end - end position of the part to copy
 */
bool concat_name_part(struct name *dest, struct name *src, int begin, int end)
{
    assert(dest != src);
    assert(begin <= end && begin >= 0 && end <= src->length);
    if (dest == src) {
        return false;
    }
    return concat_name_str_part(dest, get_buffer_of_name(src) + begin,
        get_buffer_of_name(src) + end);
}

bool concat_name_str_part(struct name *dest, const char *begin, const char *end)
{
    int newlen;

    assert(begin <= end);
    newlen = dest->length + end - begin;
    if (!expand_name(dest, newlen)) {
        return false;
    }
    strncpy(get_buffer_of_name(dest) + dest->length, begin, end - begin);
    get_buffer_of_name(dest)[newlen] = '\0';
    dest->length = newlen;
    return true;
}

/**
 *  trunc_name:
 *      Truncates a name to a new length.
 */
void trunc_name(struct name *dest, int new_len)
{
    assert(0 <= new_len && new_len <= dest->length);
    get_buffer_of_name(dest)[new_len] = '\0';
    dest->length = new_len;
}

bool expand_name(struct name *name, int expected_len)
{
    int newsize; /* new size */
    char *newbuf; /* new buffer */

    /* NOTE: expected_len + 1: here `+ 1' is for the null character. */
    if (expected_len + 1 > name->size) {
        newsize = get_min_power_of_two_ge(expected_len + 1 + NAME_PRIMBUFSIZE);
        newbuf = (char *)realloc(name->second_buf, sizeof(char) * newsize);
        if (newbuf == NULL) {
            add_error_message("Failed to malloc in expand_name.\n");
            return false;
        }
        if (name->second_buf == NULL) {
            strcpy(newbuf, name->prim_buf);
        }
        name->second_buf = newbuf;
        name->size = newsize;
    }
    return true;
}

void clear_name(struct name *name)
{
    destroy_name(name);
    init_name(name);
}

void destroy_name(struct name *name)
{
    if (name == NULL) {
        return;
    }
    if (name->destroyed) {
        return;
    }
    if (name->second_buf != NULL) {
        free(name->second_buf);
    }
    name->destroyed = true;
}

